import inspect
from unittest import TestCase

from tinkerforge2mqtt.mapping.value_map import HaValueInfo, ValueMap


DEVICE_CLASSES = {
    'date',
    'enum',
    'timestamp',
    'apparent_power',
    'aqi',
    'atmospheric_pressure',
    'battery',
    'carbon_monoxide',
    'carbon_dioxide',
    'current',
    'data_rate',
    'data_size',
    'distance',
    'duration',
    'energy',
    'energy_storage',
    'frequency',
    'gas',
    'humidity',
    'illuminance',
    'irradiance',
    'moisture',
    'monetary',
    'nitrogen_dioxide',
    'nitrogen_monoxide',
    'nitrous_oxide',
    'ozone',
    'ph',
    'pm1',
    'pm10',
    'pm25',
    'power_factor',
    'power',
    'precipitation',
    'precipitation_intensity',
    'pressure',
    'reactive_power',
    'signal_strength',
    'sound_pressure',
    'speed',
    'sulphur_dioxide',
    'temperature',
    'volatile_organic_compounds',
    'volatile_organic_compounds_parts',
    'voltage',
    'volume',
    'volume_storage',
    'volume_flow_rate',
    'water',
    'weight',
    'wind_speed',
}


class ValueMapTestCase(TestCase):
    def test_valuemap(self):
        def is_ha_value_info(value):
            return isinstance(value, HaValueInfo)

        for name, value in inspect.getmembers(ValueMap, is_ha_value_info):
            with self.subTest(name=name):
                if device_class := value.device_class:
                    self.assertIn(device_class, DEVICE_CLASSES)



import logging

from tinkerforge.bricklet_industrial_dual_relay import BrickletIndustrialDualRelay
from tinkerforge.bricklet_motion_detector_v2 import BrickletMotionDetectorV2
from tinkerforge.bricklet_temperature_v2 import BrickletTemperatureV2

from tinkerforge2mqtt.device_map import register_map_class
from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase, print_exception_decorator
from tinkerforge2mqtt.mapping.value_map import ValueMap


logger = logging.getLogger(__name__)


@register_map_class()
class BrickletTemperatureV2Mapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricks/HATZero_Brick_Python.html

    device_identifier = BrickletTemperatureV2.DEVICE_IDENTIFIER

    def __init__(self, device: BrickletTemperatureV2, ha_value_callback):
        self.device = device
        self.ha_value_callback = ha_value_callback

    def register_callbacks(self):
        self.device.set_temperature_callback_configuration(
            period=1000,  # 1000ms == 1s
            value_has_to_change=False,
            option=BrickletTemperatureV2.THRESHOLD_OPTION_OFF,
            min=-999,
            max=999,
        )
        self.device.register_callback(self.device.CALLBACK_TEMPERATURE, self.callback_temperature)

    @print_exception_decorator
    def callback_temperature(self, value):
        logger.debug(f'Temperature: {value / 100}°C (UID: {self.device.uid_string})')
        self.ha_value_callback(
            device_mapper=self,
            value=value / 100,
            ha_value_info=ValueMap.TEMPERATURE,
        )


@register_map_class()
class BrickletIndustrialDualRelayMapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricklets/IndustrialDualRelay_Bricklet_Python.html

    device_identifier = BrickletIndustrialDualRelay.DEVICE_IDENTIFIER

    def __init__(self, device: BrickletIndustrialDualRelay, ha_value_callback):
        self.device = device
        self.ha_value_callback = ha_value_callback

    def register_callbacks(self):
        self.device.register_callback(self.device.CALLBACK_MONOFLOP_DONE, self.callback_monoflop_done)

    @print_exception_decorator
    def callback_monoflop_done(self, value):
        logger.debug(f'Temperature: {value / 100}°C (UID: {self.device.uid_string})')
        self.ha_value_callback(
            device_mapper=self,
            value=value / 100,
            ha_value_info=ValueMap.TEMPERATURE,
        )


@register_map_class()
class BrickletMotionDetectorV2Mapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricklets/MotionDetectorV2_Bricklet_Python.html

    device_identifier = BrickletMotionDetectorV2.DEVICE_IDENTIFIER

    def __init__(self, device: BrickletMotionDetectorV2, ha_value_callback):
        self.device = device
        self.ha_value_callback = ha_value_callback

    @print_exception_decorator
    def poll(self):
        super().poll()
        sensitivity = self.device.get_sensitivity()
        self.ha_value_callback(
            device_mapper=self,
            value=sensitivity,
            ha_value_info=ValueMap.PIR_SENSOR_SENSITIVITY,
        )

    def register_callbacks(self):
        self.device.register_callback(self.device.CALLBACK_MOTION_DETECTED, self.callback_motion_detected)
        self.device.register_callback(self.device.CALLBACK_DETECTION_CYCLE_ENDED, self.callback_detection_cycle_ended)

    @print_exception_decorator
    def callback_motion_detected(self, value):
        logger.debug(f'Motion detected: {value=} (UID: {self.device.uid_string})')
        self.ha_value_callback(
            device_mapper=self,
            value=value,
            ha_value_info=ValueMap.MOTION_DETECTED,
        )

    @print_exception_decorator
    def callback_detection_cycle_ended(self, value):
        logger.debug(f'Detection Cycle Ended: {value} (UID: {self.device.uid_string})')
        self.ha_value_callback(
            device_mapper=self,
            value=value,
            ha_value_info=ValueMap.DETECTION_CYCLE_ENDED,
        )



import logging

from tinkerforge.brick_hat_zero import BrickHATZero

from tinkerforge2mqtt.device_map import register_map_class
from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase, print_exception_decorator
from tinkerforge2mqtt.mapping.value_map import ValueMap


logger = logging.getLogger(__name__)


@register_map_class()
class BrickHATZeroMapper(DeviceMapBase):
    # https://www.tinkerforge.com/de/doc/Software/Bricks/HATZero_Brick_Python.html

    device_identifier = BrickHATZero.DEVICE_IDENTIFIER

    def __init__(self, device: BrickHATZero, ha_value_callback):
        self.device = device
        self.ha_value_callback = ha_value_callback

    def register_callbacks(self):
        self.device.set_usb_voltage_callback_configuration(
            period=1000,  # 1000ms == 1s
            value_has_to_change=False,
            option='x',  # Threshold is turned off
            min=0,
            max=999,
        )
        self.device.register_callback(self.device.CALLBACK_USB_VOLTAGE, self.callback_usb_voltage)

    @print_exception_decorator
    def callback_usb_voltage(self, value):
        logger.debug(f'USB Voltage: {value / 1000}V (UID: {self.device.uid_string})')
        self.ha_value_callback(
            device_mapper=self,
            value=value / 1000,
            ha_value_info=ValueMap.USB_VOLTAGE,
        )



import collections
import logging
import queue
import time

from ha_services.mqtt4homeassistant.converter import values2mqtt_payload
from ha_services.mqtt4homeassistant.data_classes import HaValue, HaValues
from ha_services.mqtt4homeassistant.mqtt import HaMqttPublisher
from tinkerforge.device_factory import get_device_class
from tinkerforge.ip_connection import Device, Error, IPConnection

import tinkerforge2mqtt
from tinkerforge2mqtt.device_map import map_registry
from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase
from tinkerforge2mqtt.mapping.device_value import DeviceValue
from tinkerforge2mqtt.mapping.value_map import HaValueInfo


logger = logging.getLogger(__name__)


class DevicesHandler:
    def __init__(
        self,
        ipcon: IPConnection,
        *,
        publisher: HaMqttPublisher | None = None,
        name_prefix: str = '',
        update_interval: int = 2,
    ):
        self.ipcon = ipcon
        self.publisher = publisher
        self.name_prefix = name_prefix
        self.update_interval = update_interval

        self._map_instances = {}
        self.ha_value_queue = queue.Queue()
        self.next_update = time.monotonic() + self.update_interval

    def ha_value_callback(self, *, device_mapper: DeviceMapBase, value, ha_value_info: HaValueInfo):
        logger.debug(f'HA value callback: {device_mapper}: {value=} {ha_value_info=}')
        device_value = DeviceValue(
            device_mapper=device_mapper,
            value=value,
            info=ha_value_info,
        )
        logger.info(f'{device_value=}')
        self.ha_value_queue.put(device_value)

        if time.monotonic() > self.next_update:
            self.push_values()
            self.next_update = time.monotonic() + self.update_interval

    def push_values(self):
        device_ha_values = collections.defaultdict(list)
        while not self.ha_value_queue.empty():
            device_value: DeviceValue = self.ha_value_queue.get()

            device_mapper: DeviceMapBase = device_value.device_mapper

            ha_value_info: HaValueInfo = device_value.info
            ha_value: HaValue = ha_value_info.get_ha_value(device_value.value)

            device_ha_values[device_mapper].append(ha_value)

        for device_mapper, values in device_ha_values.items():
            device: Device = device_mapper.device
            uid_string = device.uid_string
            device_name = device.DEVICE_DISPLAY_NAME
            device_name = f'{device_name} ({uid_string})'

            ha_values = HaValues(
                device_name=device_name,
                values=values,
            )
            logger.info(f'Send {len(values)} values for {device_name} to HomeAssistant')

            # Create Payload:
            ha_mqtt_payload = values2mqtt_payload(
                values=ha_values,
                name_prefix=self.name_prefix,
                device_extra_info={
                    'manufacturer': 'Tinkerforge',
                    'sw_version': device_mapper.get_sw_version(),
                },
                origin={
                    'name': 'tinkerforge2mqtt',
                    'sw_version': tinkerforge2mqtt.__version__,
                    'support_url': 'https://pypi.org/project/tinkerforge2mqtt/',
                },
            )

            # Send vial MQTT to HomeAssistant:
            self.publisher.publish2homeassistant(ha_mqtt_payload=ha_mqtt_payload)

    def __call__(
        self,
        uid,
        connected_uid,
        position,
        hardware_version,
        firmware_version,
        device_identifier,
        enumeration_type,
    ):
        if enumeration_type == IPConnection.ENUMERATION_TYPE_DISCONNECTED:
            print('Disconnected!')
            return

        if map_instance := self._map_instances.get(uid):
            # Already initialized device -> poll values
            try:
                map_instance.poll()
            except Error as err:
                logger.exception(f'Error in poll: {err}')
        else:
            # Initialize new device

            DeviceClass = get_device_class(device_identifier)
            print('_' * 80)
            print(f'{DeviceClass.DEVICE_DISPLAY_NAME} ({DeviceClass.__name__})')

            MapClass = map_registry.get_map_class(device_identifier)
            if not MapClass:
                logger.error(f'No mapper found for {DeviceClass.__name__} ({device_identifier=})')
                return

            device: Device = DeviceClass(uid=uid, ipcon=self.ipcon)
            map_instance = MapClass(device=device, ha_value_callback=self.ha_value_callback)
            map_instance.register_callbacks()
            self._map_instances[uid] = map_instance
            print(f'{map_instance=}')



import dataclasses

from ha_services.mqtt4homeassistant.data_classes import HaValue


@dataclasses.dataclass
class HaValueInfo:
    name: str
    device_class: str | None  # e.g.: "voltage" / "current" / "energy" etc.
    state_class: str | None  # e.g.: "measurement" / "total" / "total_increasing" etc.
    unit: str | None  # e.g.: "V" / "A" / "kWh" etc.

    def get_log_message(self, device, value):
        return f'{self.name}: {value} {self.unit} [{device.DEVICE_DISPLAY_NAME} ({device.uid_string})]'

    def get_ha_value(self, value) -> HaValue:
        return HaValue(
            name=self.name,
            value=value,
            device_class=self.device_class,
            state_class=self.state_class,
            unit=self.unit,
        )


class ValueMap:
    API_VERSION = HaValueInfo(
        name='API Version',
        device_class=None,
        state_class='firmware',
        unit=None,
    )
    USB_VOLTAGE = HaValueInfo(
        name='USB Voltage',
        device_class='voltage',
        state_class='measurement',
        unit='V',
    )
    TEMPERATURE = HaValueInfo(
        name='Temperature',
        device_class='temperature',
        state_class='measurement',
        unit='°C',
    )
    CHIP_TEMPERATURE = HaValueInfo(
        name='Chip Temperature',
        device_class='temperature',
        state_class='measurement',
        unit='°C',
    )
    STATUS_LED_CONFIG = HaValueInfo(
        name='Status LED Config',
        device_class='enum',
        state_class=None,
        unit=None,
    )
    MOTION_DETECTED = HaValueInfo(
        name='Motion detected',
        device_class=None,
        state_class='motion',
        unit=None,
    )
    DETECTION_CYCLE_ENDED = HaValueInfo(
        name='Detection Cycle Ended',
        device_class=None,
        state_class='motion',
        unit=None,
    )
    PIR_SENSOR_SENSITIVITY = HaValueInfo(
        name='Sensitivity of the PIR sensor',
        device_class='weight',
        state_class='measurement',
        unit='%',
    )



import dataclasses

from tinkerforge2mqtt.device_map_utils.base import DeviceMapBase
from tinkerforge2mqtt.mapping.value_map import HaValueInfo


@dataclasses.dataclass
class DeviceValue:
    device_mapper: DeviceMapBase
    value: int | float | str
    info: HaValueInfo



import abc
import logging
from functools import wraps

from rich import print  # noqa
from rich.console import Console
from tinkerforge.ip_connection import Device

from tinkerforge2mqtt.device_map_utils.generics import iter_interest_functions
from tinkerforge2mqtt.mapping.value_map import ValueMap


logger = logging.getLogger(__name__)


def print_exception_decorator(func):

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as err:
            console = Console()
            console.print_exception(show_locals=True)
            raise SystemExit from err

    return func_wrapper


class DeviceMapBase(abc.ABC):
    device_identifier: int

    @abc.abstractmethod
    def __init__(self, device: Device, ha_value_callback):
        self.device = device
        self.ha_value_callback = ha_value_callback

    def iter_known_functions(self, device: Device):
        assert (
            device.DEVICE_IDENTIFIER == self.device_identifier
        ), f'Wrong device: {device} is not {self.device_identifier}'

        yield from iter_interest_functions(device)

    @print_exception_decorator
    def poll(self):
        value = self.device.get_chip_temperature()
        logger.debug(f'{self.device.DEVICE_DISPLAY_NAME} chip temperature: {value}°C')
        self.ha_value_callback(
            device_mapper=self,
            value=value,
            ha_value_info=ValueMap.CHIP_TEMPERATURE,
        )

        if get_status_led_config := getattr(self.device, 'get_status_led_config', None):
            value = get_status_led_config()
            logger.debug(f'{self.device.DEVICE_DISPLAY_NAME} status LED config: {value}')
            self.ha_value_callback(
                device_mapper=self,
                value=value,
                ha_value_info=ValueMap.STATUS_LED_CONFIG,
            )

    @abc.abstractmethod
    def register_callbacks(self):
        pass

    def get_sw_version(self) -> str:
        api_version = self.device.get_api_version()
        sw_version = '.'.join(str(number) for number in api_version)
        return sw_version

    def __str__(self):
        return f'{self.__class__.__name__} (UID: {self.device.uid_string})'

    def __repr__(self):
        return f'<{self}>'



import logging
import time

import rich_click as click
from cli_base.cli_tools.verbosity import OPTION_KWARGS_VERBOSE, setup_logging
from ha_services.mqtt4homeassistant.mqtt import HaMqttPublisher
from rich import print  # noqa
from tinkerforge.ip_connection import IPConnection

from tinkerforge2mqtt.cli_app import cli
from tinkerforge2mqtt.cli_app.settings import get_user_settings
from tinkerforge2mqtt.device_registry.devices_handler import DevicesHandler
from tinkerforge2mqtt.user_settings import UserSettings


logger = logging.getLogger(__name__)


@cli.command()
@click.option('-v', '--verbosity', **OPTION_KWARGS_VERBOSE | {'default': 0})
def publish_loop(verbosity: int):
    """
    Discover Victron devices with Instant Readout
    """
    setup_logging(verbosity=verbosity)
    user_settings: UserSettings = get_user_settings(verbosity=verbosity)

    publisher = HaMqttPublisher(
        settings=user_settings.mqtt,
        verbosity=verbosity - 1,
        config_count=1,  # Send every time the config
    )

    # https://www.tinkerforge.com/en/doc/Software/IPConnection_Python.html
    ipcon = IPConnection()
    connect_kwargs = dict(
        host=user_settings.host,
        port=user_settings.port,
    )
    print(f'Connecting to {connect_kwargs}')
    ipcon.connect(**connect_kwargs)

    devices_handler = DevicesHandler(
        ipcon,
        publisher=publisher,
        name_prefix=user_settings.mqtt_payload_prefix,
        update_interval=user_settings.update_interval,
    )

    ipcon.register_callback(IPConnection.CALLBACK_ENUMERATE, devices_handler)

    print('Aborting with Ctrl-C !')
    while True:
        try:
            ipcon.enumerate()
            time.sleep(user_settings.enumerate_sleep)
        except Exception as err:
            logger.exception('Exception in enumerate loop: %s', err)
            break
        except KeyboardInterrupt:
            logger.info('Keyboard interrupt')
            ipcon.disconnect()
            break
